import std/[atomics, bitops, critbits, strformat, strutils, strmisc, math, re]
import bio
import itertools

type
    BitShiftType = range[0 .. 64]
    ArcticTypeKind* = enum
        BYTE, SHORT, INT, WORD, DOUBLE

    ArcticType* {.union.} = object
        b: int8
        s: int16
        i: int32
        w: int64 # TODO: swap i and w to match instruction naming -.-
        u: uint64
        d: float64
        f: ArcticSymbol # also used for pointers
        # TODO: Eventually replace f with p/f and each is just a uint64
        #       Then load all memory into one long seq instead of 1 seq per
        #       section. Probably actually have +ve be data memory, and -ve be
        #       code memory (with sections mapping to ranges)

    ArcticVariableIndex* = enum
        VARIABLE_A, VARIABLE_B, VARIABLE_C,
        VARIABLE_I, VARIABLE_N,
        VARIABLE_X, VARIABLE_Y, VARIABLE_Z

    ArcticSymbol* = tuple
        section: string
        index: int

    ArcticImmediateKind* = enum
        PLAIN, VARIABLE, INTEGER, NUMBER, SYMBOL

    ArcticImmediate* = object
        case kind: ArcticImmediateKind
            of PLAIN: nil
            of VARIABLE: v: ArcticVariableIndex
            of INTEGER: i: int64
            of NUMBER: n: float64
            of SYMBOL: s: string

    ArcticOperation* = object
        code: char
        immediate: ArcticImmediate

    ArcticCode* = seq[ArcticOperation]
    ArcticMemory* = seq[uint8]

    ArcticSection* = object
        case iscode: bool
            of true: code: ArcticCode
            of false: data: ArcticMemory
    ArcticStack* = seq[ArcticType]

    ArcticState* = object
        symbols: CritBitTree[ArcticSymbol] # as name or section _name
        sections: CritBitTree[ArcticSection]
        stack: ArcticStack
        pc: ArcticSymbol
        registers: array[ArcticVariableIndex, ArcticType]

    ArcticStepResult* = enum
        CONTINUE, BREAKPOINT, ERROR

    ArcticBuiltin* = proc (state: var ArcticState): ArcticStepResult

var
    next_malloc_id: int = 1

const
    ImmediateOps: set[char] = {'0', 'E', 'F', 'G', 'H', 'L', 'M', 'P', 'Q', 'd', 'o', 'u', '<', '{', '=', '}', '>', ',', ';', '[', ']'}
    PlainOps : set[char] = { '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'I', 'J', 'K', 'N', 'O', 'R', 'S', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'p', 'q', 'r', 's', 't', 'v', 'w', 'x', 'y', 'z', '_', '.', '+', '-', '*', '/', '%', '\\', '|', '&', '^', '!', '?', '(', ')', '`' }

proc `$`*(variable: ArcticVariableIndex): string =
    case variable:
        of VARIABLE_A: return "A"
        of VARIABLE_B: return "B"
        of VARIABLE_C: return "C"
        of VARIABLE_I: return "I"
        of VARIABLE_N: return "N"
        of VARIABLE_X: return "X"
        of VARIABLE_Y: return "Y"
        of VARIABLE_Z: return "Z"

proc `$`*(op: ArcticOperation): string =
    result &= op.code

    case op.immediate.kind:
        of PLAIN: discard
        of VARIABLE: result &= $op.immediate.v
        of INTEGER: result &= $op.immediate.i
        of NUMBER: result &= $op.immediate.n
        of SYMBOL: result &= "$" & $op.immediate.s

    result &= " "

proc `$`*(state: ArcticState): string =
    result &= "Registers:\n"
    for idx in ArcticVariableIndex:
        result &= &" {idx} = {state.registers[idx]}\n"

    result &= "Stack:\n"

    for item in state.stack:
        result &= &" {item}\n"

    for (name, section) in state.sections.pairs:
        result &= &"\n\"{name}\": "
        if section.iscode:
            result &= "code \n"
            for (i, m) in section.code.pairs:
                for (label, location) in state.symbols.pairs:
                    if location.section == name and location.index == i:
                        result &= &":{label} "
                if state.pc.section == name and state.pc.index == i:
                    result &= ">"
                result &= $m
        else:
            result &= "\n"
            for (i, m) in section.data.pairs:
                for (label, location) in state.symbols.pairs:
                    if location.section == name and location.index == i:
                        result &= &":{label} "
                result &= &"{m} "

proc op(code: char, immediate: ArcticImmediate = ArcticImmediate(kind: PLAIN)): ArcticOperation =
    return ArcticOperation(code: code, immediate: immediate)

proc add(section: var ArcticSection, code: char) =
    if section.iscode:
        section.code.add code.op
    else:
        section.data.add code.uint8

proc tovar(code: char): ArcticVariableIndex =
    case code.toLowerAscii:
        of 'a': VARIABLE_A
        of 'b': VARIABLE_B
        of 'c': VARIABLE_C
        of 'i': VARIABLE_I
        of 'n': VARIABLE_N
        of 'x': VARIABLE_X
        of 'y': VARIABLE_Y
        of 'z': VARIABLE_Z
        else: raise new ValueError

proc add(section: var ArcticSection, code: char, immstr: string, secname: string) =
    if section.iscode:
        if immstr.match(re"^\d+$"):
            section.code.add code.op(ArcticImmediate(kind: INTEGER, i: immstr.parseInt))
        elif immstr.match(re"^[AaBbCcIiNnXxYyZz]$"):
            section.code.add code.op(ArcticImmediate(kind: VARIABLE, v: immstr[0].tovar))
        elif immstr[0] == '$':
            if immstr.len == 1 or immstr[1] == '_':
                section.code.add code.op(ArcticImmediate(kind: SYMBOL, s: secname & " " & immstr))
            else:
                section.code.add code.op(ArcticImmediate(kind: SYMBOL, s: immstr))
        else:
            section.code.add code.op(ArcticImmediate(kind: NUMBER, n: immstr.parseFloat))

proc len(section: ArcticSection): int =
    if section.iscode:
        return section.code.len
    else:
        return section.data.len

func grabnum(data: string): (string, string) =
    ## Returns the numeric literal at the start of data, then the rest of data
    var i = 0
    while i < data.len and data[i] in "0123456789.":
        result[0] &= data[i]
        i += 1

    result[1] = data[i..^1]

proc insert(section: var ArcticSection, data: string) =
    if data.len == 0:
        return

    if section.iscode:
        return

    template insertInt(itype, utype, endianness) =
        let (text, rest) = data[1..^1].grabnum

        if text[0] == '-':
            let val: itype = itype(-text[1..^1].parseBiggestInt)
            section.data.add cast[utype](val).serialize(endianness)
        else:
            let val: utype = utype(text.parseBiggestUInt)
            section.data.add val.serialize(endianness)

        section.insert(rest)

    case data[0]:
        of 'i': insertInt(int64, uint64, littleEndian)
        of 'I': insertInt(int64, uint64, bigEndian)
        of 'w': insertInt(int32, uint32, littleEndian)
        of 'W': insertInt(int32, uint32, bigEndian)
        of 's': insertInt(int16, uint16, littleEndian)
        of 'S': insertInt(int16, uint16, bigEndian)
        of 'b', 'B': # bytes don't have an endianness
            let (text, rest) = data[1..^1].grabnum

            if text[0] == '-':
                let val: int8 = int8(text[1..^1].parseBiggestInt)
                section.data.add cast[uint8](val)
            else:
                let val: uint8 = uint8(text.parseBiggestUInt)
                section.data.add val

        of 'f':
            let (text, rest) = data[1..^1].grabnum
            section.data.add cast[uint64](text.parseFloat).serialize(littleEndian)
            section.insert(rest)
        of 'F':
            let (text, rest) = data[1..^1].grabnum
            section.data.add cast[uint64](text.parseFloat).serialize(bigEndian)
            section.insert(rest)

        of 'x': # hexadecimal byte constant
            let (hex, x, rest) = data[1..^1].partition("x")
            for pair in hex.chunked(2):
                section.data.add uint8(parseHexInt(pair.join))
            section.insert(rest)

        of '"': # utf-8 string constant with escapes
            let strlen = data.matchLen(re"^""([^""]|\\"")*""")

            if strlen < 0:
                return # TODO: couldn't find a string, what to do?

            for c in data[1..^(strlen-2)]:
                section.data.add c.uint8

            section.insert(data[strlen..^1])

        else:
            section.insert(data[1..^1])

func load*(code: string): ArcticState =
    var
        section: string = ""
        token: string = ""

    result.pc = ("", 0)
    result.sections[section] = ArcticSection(iscode: true, code: @[])

    for next in code:
        debug_echo "token: ", token, ", next: ", next
        if token.len == 0: # initial state
            case next:
                of ImmediateOps, '#', ':', '\"', '\'':
                    token &= next
                of PlainOps, '\n':
                    result.sections[section].add next
                else:
                    discard
        else:
            case token[0]:
                of '#': # comment
                    if next == '\n':
                        token = ""
                        result.sections[section].add next

                of '"': # section switch
                    if next == '"' and token[^1] != '\\':
                        section = token[1..^1]
                        if section notin result.sections:
                            if section.toLowerAscii in @["code", "macros", ""]:
                                result.sections[section] = ArcticSection(iscode: true, code: @[])
                            else:
                                result.sections[section] = ArcticSection(iscode: false, data: @[])
                        token = ""
                    else:
                        token.add next

                of ':': # label name
                    if next == ' ' or next == '\n':
                        let idx = if token[1] == '_': section & " " & token[1..^1] else: token[1..^1]
                        result.symbols[idx] = (section: section, index: result.sections[section].len)
                        token = ""
                    else:
                        token.add next

                of '\'': # data injection
                    if next == '\'':
                        var
                            quoted = false
                            escaped = false
                        for c in token:
                            if c == '"':
                                if not escaped:
                                    quoted = not quoted
                                escaped = false
                            elif c == '\\':
                                escaped = true
                            else:
                                escaped = false

                        if quoted or escaped:
                            token.add next
                        else:
                            result.sections[section].insert token[1..^1]
                            token = ""
                    else:
                        token.add next

                of ImmediateOps:
                    if next == ' ':
                        result.sections[section].add(token[0], token[1..^1], section)
                        token = ""
                    else:
                        token.add next

                else:
                    discard

proc `[]`(state: ArcticState, symbol: ArcticSymbol): uint8 =
    assert symbol.section in state.sections
    let section = state.sections[symbol.section]

    if section.iscode:
        return section.code[symbol.index].code.uint8
    else:
        return section.data[symbol.index]

proc `[]=`(state: var ArcticState, symbol: ArcticSymbol, value: uint8) =
    if not (symbol.section in state.sections):
        state.sections[symbol.section] = ArcticSection(iscode: false, data: @[])
        while state.sections[symbol.section].len <= symbol.index:
            state.sections[symbol.section].data.add 0
        state.sections[symbol.section].data[symbol.index] = value
    else:
        let section = state.sections[symbol.section]

        while state.sections[symbol.section].data.len <= symbol.index:
            state.sections[symbol.section].data.add 0
        state.sections[symbol.section].data[symbol.index] = value

proc `+`(left: ArcticSymbol, offset: int): ArcticSymbol = (left.section, left.index + offset)

proc step*(state: var ArcticState, builtins: CritBitTree[ArcticBuiltin]): ArcticStepResult =
    let codesec = state.sections[state.pc.section]

    if not codesec.iscode:
        return ERROR

    if state.pc.index >= codesec.code.len:
        return ERROR

    let op = codesec.code[state.pc.index]
    state.pc.index.inc

    case op.code:
        # Memory Ops -----------------------------------------------------------
        of '1': # LDAB
            let p = state.stack.pop.f
            state.stack.add ArcticType(w: state[p].int64)

        of '2': # LDAS
            let
                p = state.stack.pop.f
                a = state[p].int64
                b = state[p+1].int64
            if state.pc.section == "CODE": # bigendian
                state.stack.add ArcticType(w: a or b shl 8)
            else:
                state.stack.add ArcticType(w: b or a shl 8)

        of '3': # LDAI
            let
                p = state.stack.pop.f
                a = state[p].int64
                b = state[p+1].int64
                c = state[p+2].int64
                d = state[p+3].int64
            if state.pc.section == "CODE": # bigendian
                state.stack.add ArcticType(w: d or c shl 8 or b shl 16 or a shl 24)
            else:
                state.stack.add ArcticType(w: a or b shl 8 or c shl 16 or d shl 24)

        of '4': # LDAW
            let
                p = state.stack.pop.f
                a = state[p].int64
                b = state[p+1].int64
                c = state[p+2].int64
                d = state[p+3].int64
                e = state[p+4].int64
                f = state[p+5].int64
                g = state[p+6].int64
                h = state[p+7].int64
            if state.pc.section == "CODE": # bigendian
                state.stack.add ArcticType(w: h or g shl 8 or f shl 16 or e shl 24 or d shl 32 or c shl 40 or b shl 48 or a shl 56)
            else:
                state.stack.add ArcticType(w: a or b shl 8 or c shl 16 or d shl 24 or e shl 32 or f shl 40 or g shl 48 or h shl 56)

        of 'o': # LDOB
            assert op.immediate.kind == INTEGER
            let p = state.stack.pop.f
            state.stack.add ArcticType(w: state[p+op.immediate.i].int64)

        of 'H': # LDOS
            assert op.immediate.kind == INTEGER
            let
                p = state.stack.pop.f + op.immediate.i
                a = state[p].int64
                b = state[p+1].int64
            if state.pc.section == "CODE": # bigendian
                state.stack.add ArcticType(w: a or b shl 8)
            else:
                state.stack.add ArcticType(w: b or a shl 8)

        of 'G': # LDOI
            let
                p = state.stack.pop.f + op.immediate.i
                a = state[p].int64
                b = state[p+1].int64
                c = state[p+2].int64
                d = state[p+3].int64
            if state.pc.section == "CODE": # bigendian
                state.stack.add ArcticType(w: d or c shl 8 or b shl 16 or a shl 24)
            else:
                state.stack.add ArcticType(w: a or b shl 8 or c shl 16 or d shl 24)

        of 'Q': # LDOW
            let
                p = state.stack.pop.f
                a = state[p].int64
                b = state[p+1].int64
                c = state[p+2].int64
                d = state[p+3].int64
                e = state[p+4].int64
                f = state[p+5].int64
                g = state[p+6].int64
                h = state[p+7].int64
            if state.pc.section == "CODE": # bigendian
                state.stack.add ArcticType(w: h or g shl 8 or f shl 16 or e shl 24 or d shl 32 or c shl 40 or b shl 48 or a shl 56)
            else:
                state.stack.add ArcticType(w: a or b shl 8 or c shl 16 or d shl 24 or e shl 32 or f shl 40 or g shl 48 or h shl 56)

        of '5': # STAB
            let
                x = cast[uint8](state.stack.pop.b)
                p = state.stack.pop.f
            state[p] = x

        of '6': # STAS
            let
                x = cast[uint16](state.stack.pop.s)
                p = state.stack.pop.f
            if state.pc.section == "CODE": # bigendian
                state[p] = uint8(x and 0xFF)
                state[p+1] = uint8(x shr 8)
            else:
                state[p] = uint8(x shr 8)
                state[p+1] = uint8(x and 0xFF)

        of '7': # STAI
            let
                x = cast[uint32](state.stack.pop.i)
                p = state.stack.pop.f
            if state.pc.section == "CODE": # bigendian
                state[p] = uint8(x and 0xFF)
                state[p+1] = uint8((x shr 8) and 0xFF)
                state[p+2] = uint8((x shr 16) and 0xFF)
                state[p+3] = uint8((x shr 24) and 0xFF)
            else:
                state[p] = uint8((x shr 24) and 0xFF)
                state[p+1] = uint8((x shr 16) and 0xFF)
                state[p+2] = uint8((x shr 8) and 0xFF)
                state[p+3] = uint8(x and 0xFF)

        of '8': # STAW
            let
                x = cast[uint64](state.stack.pop.w)
                p = state.stack.pop.f
            if state.pc.section == "CODE": # bigendian
                state[p] = uint8(x and 0xFF)
                state[p+1] = uint8((x shr 8) and 0xFF)
                state[p+2] = uint8((x shr 16) and 0xFF)
                state[p+3] = uint8((x shr 24) and 0xFF)
                state[p+4] = uint8((x shr 32) and 0xFF)
                state[p+5] = uint8((x shr 40) and 0xFF)
                state[p+6] = uint8((x shr 48) and 0xFF)
                state[p+7] = uint8((x shr 56) and 0xFF)
            else:
                state[p] = uint8((x shr 56) and 0xFF)
                state[p+1] = uint8((x shr 48) and 0xFF)
                state[p+2] = uint8((x shr 40) and 0xFF)
                state[p+3] = uint8((x shr 32) and 0xFF)
                state[p+4] = uint8((x shr 24) and 0xFF)
                state[p+5] = uint8((x shr 16) and 0xFF)
                state[p+6] = uint8((x shr 8) and 0xFF)
                state[p+7] = uint8(x and 0xFF)

        of 'D': # MCLR
            let
                x = state.stack.pop.w
                p = state.stack.pop.f
            for i in 0 .. x:
                state[p+i] = 0

        of 'd': # MSET
            assert op.immediate.kind == INTEGER
            let
                x = state.stack.pop.w
                p = state.stack.pop.f
                b = op.immediate.i.uint8
            for i in 0 .. x:
                state[p+i] = b

        of 'K': # MCPY
            let
                x = state.stack.pop.w
                q = state.stack.pop.f
                p = state.stack.pop.f
            for i in 0 .. x:
                state[q+i] = state[p+i]

        of 'J': # MOFF
            let
                x = state.stack.pop.w
                p = state.stack.pop.f
            state.stack.add ArcticType(f: (section: p.section, index: p.index + 1))

        of 'M': # MALL
            assert op.immediate.kind == INTEGER
            let
                top = state.stack.pop
                n = op.immediate.i

            if n == 0:
                if top.w != 0:
                    state.sections.excl top.f.section
                    state.stack.add ArcticType(w: 0)
            elif n > 0:
                if top.w == 0:
                    state.sections.incl("allocated " & $next_malloc_id, ArcticSection(iscode: false, data: @[]))
                    inc next_malloc_id
                else:
                    discard # we use dynamic memory, realloc is a no-op
            else:
                discard # realloc negative is a nop


        of 'R': # REAL
            let
                top = state.stack.pop
                x = state.stack.pop.i

            if x == 0:
                if top.w != 0:
                    state.sections.excl top.f.section
                    state.stack.add ArcticType(w: 0)
            elif x > 0:
                if top.w == 0:
                    state.sections.incl("allocated " & $next_malloc_id, ArcticSection(iscode: false, data: @[]))
                    inc next_malloc_id
                else:
                    discard # we use dynamic memory, realloc is a no-op
            else:
                discard # realloc negative is a nop


        of '9': # CASS
            # TODO: Actually CAS? need to use Atomic[T] for that...
            #       implementation would probably require an "atomic" memory
            #       section
            let
                y = state.stack.pop.w
                x = state.stack.pop.w
                p = state.stack.pop.f
                a = state[p].int64
                b = state[p+1].int64
                c = state[p+2].int64
                d = state[p+3].int64
                e = state[p+4].int64
                f = state[p+5].int64
                g = state[p+6].int64
                h = state[p+7].int64

            var old: int64
            if state.pc.section == "CODE": # bigendian
                old = h or g shl 8 or f shl 16 or e shl 24 or d shl 32 or c shl 40 or b shl 48 or a shl 56
            else:
                old = a or b shl 8 or c shl 16 or d shl 24 or e shl 32 or f shl 40 or g shl 48 or h shl 56

            if x == old:
                state.stack.add ArcticType(i: 1)
                if state.pc.section == "CODE": # bigendian
                    state[p] = uint8(y and 0xFF)
                    state[p+1] = uint8((y shr 8) and 0xFF)
                    state[p+2] = uint8((y shr 16) and 0xFF)
                    state[p+3] = uint8((y shr 24) and 0xFF)
                    state[p+4] = uint8((y shr 32) and 0xFF)
                    state[p+5] = uint8((y shr 40) and 0xFF)
                    state[p+6] = uint8((y shr 48) and 0xFF)
                    state[p+7] = uint8((y shr 56) and 0xFF)
                else:
                    state[p] = uint8((y shr 56) and 0xFF)
                    state[p+1] = uint8((y shr 48) and 0xFF)
                    state[p+2] = uint8((y shr 40) and 0xFF)
                    state[p+3] = uint8((y shr 32) and 0xFF)
                    state[p+4] = uint8((y shr 24) and 0xFF)
                    state[p+5] = uint8((y shr 16) and 0xFF)
                    state[p+6] = uint8((y shr 8) and 0xFF)
                    state[p+7] = uint8(y and 0xFF)
            else:
                state.stack.add ArcticType(w: 0)


        # Math Ops -----------------------------------------------------------
        of '+': # IADD
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: x + y)

        of '-': # ISUB
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: x - y)

        of '*': # IMUL
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: x * y)

        of '/': # IDIV
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: int(x / y))

        of '%': # IMOD
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: x.floorMod(y))

        of '\\': # IREM
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: x mod y)

        of '_': # INEG
            let x = state.stack.pop.w
            state.stack.add ArcticType(w: -x)

        of '|': # IABS
            let x = state.stack.pop.w
            state.stack.add ArcticType(w: x.abs)

        of 's': # FADD
            let
                b = state.stack.pop.d
                a = state.stack.pop.d

            state.stack.add ArcticType(d: a + b)

        of 'm': # FSUB
            let
                b = state.stack.pop.d
                a = state.stack.pop.d

            state.stack.add ArcticType(d: a - b)

        of 'p': # FMUL
            let
                b = state.stack.pop.d
                a = state.stack.pop.d

            state.stack.add ArcticType(d: a * b)

        of 'q': # FDIV
            let
                b = state.stack.pop.d
                a = state.stack.pop.d

            state.stack.add ArcticType(d: a / b)

        of 'f': # FMOD
            let
                b = state.stack.pop.d
                a = state.stack.pop.d

            state.stack.add ArcticType(d: floorMod(a, b))

        of 'r': # FREM
            let
                b = state.stack.pop.d
                a = state.stack.pop.d

            state.stack.add ArcticType(d: a mod b)

        of 'j': # FNEG
            let
                a = state.stack.pop.d

            state.stack.add ArcticType(d: -a)

        of 'v': # FABS
            let
                a = state.stack.pop.d

            state.stack.add ArcticType(d: a.abs)

        of 'V': # BIOR
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: x or y)

        of '&': # BAND
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: x and y)

        of '^': # BXOR
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: x xor y)

        of 'l': # BITC
            let
                y = state.stack.pop.w
                x = state.stack.pop.w

            state.stack.add ArcticType(w: x and (not y))

        of '!': # BNOT
            let x = state.stack.pop.w

            state.stack.add ArcticType(w: not x)

        of 'u': # USHR
            let x = state.stack.pop.w
            var i: int64
            case op.immediate.kind:
                of INTEGER:
                    i = op.immediate.i
                of VARIABLE:
                    let v = op.immediate.v
                    i = state.registers[v].w
                else:
                    assert false

            if i > 0:
                state.stack.add ArcticType(w: x shr i)
            else:
                state.stack.add ArcticType(w: x shl (-i))

        of '[': # ROTR
            let x = state.stack.pop.u
            var i: int64
            case op.immediate.kind:
                of INTEGER:
                    i = op.immediate.i
                of VARIABLE:
                    let v = op.immediate.v
                    i = state.registers[v].w
                else:
                    assert false

            if i > 0:
                state.stack.add ArcticType(u: rotateRightBits(x, i))
            else:
                state.stack.add ArcticType(u: rotateLeftBits(x, -i))

        of ']': # BSHR
            let x = state.stack.pop.u
            var i: int64
            case op.immediate.kind:
                of INTEGER:
                    i = op.immediate.i
                of VARIABLE:
                    let v = op.immediate.v
                    i = state.registers[v].w
                else:
                    assert false

            if i > 0:
                state.stack.add ArcticType(u: rotateRightBits(x, i))
            else:
                state.stack.add ArcticType(u: rotateLeftBits(x, -i))

        of '?': # CMPI
            let
                y = state.stack.pop.w
                x = state.stack.pop.w
            if x > y:
                state.stack.add ArcticType(w: 1)
            elif y > x:
                state.stack.add ArcticType(w: -1)
            else:
                state.stack.add ArcticType(w: 0)

        of '~': # CMPF
            assert op.immediate.kind == INTEGER
            let
                b = state.stack.pop.d
                a = state.stack.pop.d
            if almostEqual(a, b, op.immediate.i.Natural):
                state.stack.add ArcticType(d: 0.0)
            elif a > b:
                state.stack.add ArcticType(d: 1.0)
            else:
                state.stack.add ArcticType(d: -1.0)

        of 'U': # CMPU
            let
                y = state.stack.pop.u
                x = state.stack.pop.u
            if x > y:
                state.stack.add ArcticType(w: 1)
            elif y > x:
                state.stack.add ArcticType(w: -1)
            else:
                state.stack.add ArcticType(w: 0)

        of 'F': # BMIS
            discard # TODO: define BMIS

        # Type Conversion Ops --------------------------------------------------
        of 'E': # FTOI
            let a = state.stack.pop.d
            state.stack.add ArcticType(w: a.int64)

        of 'O': # ITOC
            let x = state.stack.pop.b
            state.stack.add ArcticType(w: x.int64)

        of 'S': # ITOS
            let x = state.stack.pop.s
            state.stack.add ArcticType(w: x.int64)

        of 'W': # ITOW
            let x = state.stack.pop.i
            state.stack.add ArcticType(w: x.int64)

        of '.': # ITOF
            let x = state.stack.pop.w
            state.stack.add ArcticType(d: x.float64)

        # Stack Ops ------------------------------------------------------------
        of 'w': # SDUP
            let x = state.stack.pop
            state.stack.add x
            state.stack.add x

        of 'k': # SPOP
            discard state.stack.pop

        of 't': # SWAP
            let
                y = state.stack.pop
                x = state.stack.pop
            state.stack.add y
            state.stack.add x

        of 'h': # OVER
            let
                y = state.stack.pop
                x = state.stack.pop
            state.stack.add x
            state.stack.add y
            state.stack.add x

        of 'e': # SROT
            let
                z = state.stack.pop
                y = state.stack.pop
                x = state.stack.pop
            state.stack.add y
            state.stack.add z
            state.stack.add x

        of 'P': # PICK
            assert op.immediate.kind == INTEGER
            let val = state.stack[^op.immediate.i]
            state.stack.add val

        of 'L': # ROLL
            assert op.immediate.kind == INTEGER
            let idx: int = state.stack.len - op.immediate.i # TODO: check for OBOEs
            state.stack.delete(idx)

        of 'T': # PUSH
            assert op.immediate.kind == INTEGER
            let
                idx: int = state.stack.len - op.immediate.i # TODO: check for OBOEs
                x = state.stack.pop
            state.stack.insert(x, idx)

        of '(': # DPTH
            state.stack.add ArcticType(w: state.stack.len)

        of ')': # PACK
            let
                x = state.stack.pop.w
                p = state.stack.pop.f
            # TODO: Can't implement this until we have unified addresses
            # Once we have the addresses, can pack x items of stack at p+...

        # Control Ops ----------------------------------------------------------
        ## TODO

    ## TODO


when isMainModule:
    let state = stdin.readAll.load
    echo state
